import functools
from typing_extensions import Self

import jax
from flax import struct
import jax.numpy as jnp

from semistaticsim.rendering.floor import split_key
from semistaticsim.groundtruth.schedule_generation import full_pattern
from semistaticsim.groundtruth.spoof_vmap import spoof_vmap

from semistaticsim.groundtruth.transmat_utils import get_scc

TIME_SCALES_MAPPING = {
    "year": 365 * 24,
    "month": 30 * 24,
    "week": 7 * 24,
    "day": 24,
    "hour": 1,
}

EPS = 0.001

@struct.dataclass
class ScheduleDistribution:
    means: jnp.array
    stds: jnp.array

    def sample_duration(self, key, i: int):
        key, rngs = split_key(key, len(self))

        def sample(k, m, s):
            # sigma2 = jnp.log(1 + (s ** 2) / (m ** 2 + 1e-8))
            # sigma = jnp.sqrt(sigma2)
            # mu = jnp.log(m + 1e-8) - sigma2 / 2
            # # Sample lognormal
            # return jnp.exp(jax.random.normal(k) * sigma + mu)
            x = jax.random.normal(key=k) * s + m
            # Skews the distribution to be non-negative
            return jnp.maximum(x, 0.0)

        # note: in "instant" mode, the returned sample will be negative
        samples = jax.vmap(sample)(rngs, self.means, self.stds)
        return jnp.take(samples, jnp.astype(i, int))

    def __len__(self):
        return self.stds.shape[0]

@struct.dataclass
class ScheduleTransition:
    mat: jnp.ndarray

    @property
    def possible_states(self):
        return jnp.arange(self.mat.shape[0])

    def transit(self, key, current: jnp.ndarray):
        current = jnp.atleast_1d(current)
        curvec = jnp.take(self.mat, jnp.astype(current, int), axis=0).squeeze()

        next_state = jax.random.choice(key, self.possible_states, shape=(1,), p=curvec)
        return next_state

    def largest_scc(self):
        return get_scc(self.mat)


@struct.dataclass
class Schedule:
    time_scale: float
    distribution: ScheduleDistribution
    transition_mat: ScheduleTransition
    subpatterns: Self
    current_pattern: int
    locations: jnp.array
    time_left_for_mode: float
    dt: float
    scales: list = struct.field(pytree_node=False)

    force_next_state: int
    safe_first_state: int

    def __len__(self):
        return len(self.distribution)

    @property
    def get_static_full_depth(self):
        return tuple([None] * len(self.scales))

    @classmethod
    def create(cls, key, dico, scales, locations, dt=1): # default dt is one hour
        time_scale_name = list(dico.keys())[0]
        time_scale = TIME_SCALES_MAPPING[time_scale_name]
        description = dico[time_scale_name]

        distribution = ScheduleDistribution(description["distribution"]["means"], description["distribution"]["stds"])
        transition_mat = ScheduleTransition(description["distribution"]["transition_mat"])
        if "locations" in description:
            subpatterns = []
        else:
            subpatterns = []
            for s in description["subpatterns"]:
                key, rng = jax.random.split(key)
                subpatterns.append(Schedule.create(rng, s, scales[1:], locations))
            if len(subpatterns) > 0:
                subpatterns = spoof_vmap(subpatterns)

        nodes_of_largest_cycle = list(transition_mat.largest_scc())

        safe_first_state = jnp.ones(1, dtype=float) * nodes_of_largest_cycle[0]

        return cls(
            time_scale=time_scale,
            distribution=distribution,
            transition_mat=transition_mat,
            subpatterns=subpatterns,
            scales=scales,
            current_pattern=-1.,
            locations=jnp.array(locations),
            time_left_for_mode=dt-EPS, # the next time we tick, we will for sure cause a transition
            dt=dt,
            force_next_state=safe_first_state,
            safe_first_state=safe_first_state,
        ).tick(key)

    def reset_subpatterns(self, depth=None):
        new_subpatterns = jax.vmap(functools.partial(Schedule.reset, depth=depth))(self=self.subpatterns)
        return new_subpatterns

    def reset(self, depth=None):
        if depth is None:
            depth = self.get_static_full_depth

        self = self.replace(force_next_state=self.safe_first_state, time_left_for_mode=-1., current_pattern=-1.)

        if len(depth) == 1:
            return self

        return self.replace(subpatterns=self.reset_subpatterns(depth=depth[1:]))


    #@jax.jit
    def tick(self, key, depth=None):
        if depth is None:
            depth = self.get_static_full_depth

        new_time_left = self.time_left_for_mode - self.dt
        time_over = jnp.astype(new_time_left <= 0, int)
        stolen_time = jnp.abs(new_time_left) * time_over

        # IF THE TIME IS OVER, COMPUTE WHAT THE NEXT MODE IS (LOOK AT THE TRANSITION MATRIX)
        key, rng = jax.random.split(key)
        DO_FORCE_NEXT_MODE = jnp.astype(self.force_next_state > -1, int) # this happens at creation, and also when a higher pattern exits (see reset)
        new_current_mode_IF_FORCE_NEXT_MODE = self.force_next_state
        self = self.replace(force_next_state=jnp.ones(1) * -1)    # no matter what, only force next state the first time
        new_current_mode_IFNOT_FORCE_NEXT_MODE = self.transition_mat.transit(rng, self.current_pattern)
        new_current_mode = new_current_mode_IFNOT_FORCE_NEXT_MODE * (1-DO_FORCE_NEXT_MODE) + new_current_mode_IF_FORCE_NEXT_MODE * DO_FORCE_NEXT_MODE
        new_current_mode = new_current_mode.squeeze()

        new_time_left = new_time_left * (1-time_over) + time_over * (self.distribution.sample_duration(key, new_current_mode) * self.time_scale - stolen_time)
        ret_current_mode = self.current_pattern * (1-time_over) + time_over * new_current_mode
        self = self.replace(current_pattern=ret_current_mode,time_left_for_mode=new_time_left)

        if len(depth) == 1:
            return self

        self_subpatterns = jax.lax.cond(time_over, lambda: self.reset_subpatterns(depth[1:]), lambda: self.subpatterns)

        key, rngs = split_key(key, len(self_subpatterns))
        new_subpatterns = jax.vmap(lambda key, x: x.tick(key, depth[1:]))(rngs, self_subpatterns)
        return self.replace(subpatterns=new_subpatterns)

    @jax.jit
    def get_current_mode(self, depth=None):
        if jnp.atleast_1d(self.dt).shape[0] != 1:
            ret = jax.vmap(lambda x:x.get_current_mode(depth))(self)
            if len(ret.shape) == 1:
                ret = ret[:,None]
            return ret

        if depth is None:
            depth = self.get_static_full_depth

        if len(depth) == 1:
            return jnp.take(self.locations, jnp.astype(self.current_pattern, int))

        ret = jax.vmap(lambda x: x.get_current_mode(depth=depth[1:]))(self.subpatterns)

        subpattern = jnp.atleast_1d(jnp.take(ret, jnp.astype(self.current_pattern, int), axis=0).squeeze())
        return jnp.concatenate([jnp.atleast_1d(self.current_pattern), subpattern]) #ret[self.current_pattern]

    @property
    def current_location(self):
        ret = jnp.atleast_2d(self.get_current_mode())
        return ret[:,-1]

    def step(self, key):
        old_self = self # noqa debug
        self = self.tick(key)
        mode = self.get_current_mode()
        return self, mode

    def make_step_func(self, scan_size=1000):
        def thunk(key, schedule):
            return jax.lax.scan(f=Schedule.step, init=schedule, xs=split_key(key, scan_size)[-1])

        print(f"Each scan will produce {scan_size} increments of {self.dt} days, which is {scan_size * self.dt} days long.")
        return thunk

if __name__ == "__main__":
    scales = ["day"]
    locations = [0,1]

    pattern = full_pattern(
        key=jax.random.PRNGKey(0),
        scales=scales, locations=locations,
        min_time_buckets=2,
        max_time_buckets=2,
        #dt=1/60
       # transition_model_mode="uniform_no_diag"
    )
    key = jax.random.PRNGKey(0)
    schedule = Schedule.create(key, pattern, scales, locations, dt=1/60)

    for i in range(1000000):
        print(schedule.current_location)
        print(schedule.time_left_for_mode)
        schedule = schedule.tick(key)
    #scan_func = schedule.make_step_func()
    #schedule, results = scan_func(key, schedule)
    # Results are the arrows of the tree
    #print(results)



import copy
import os
import sys
from typing import Union, Literal, List, Optional

from semistaticsim.groundtruth.jax_polygon import split_key
from semistaticsim.groundtruth.transmat_utils import get_scc

os.environ["JAX_PLATFORM_NAME"] = "cpu"
import jax

jax.config.update("jax_platform_name", "cpu")
from jax import numpy as jnp


TRANSITION_MODEL_LITERALS = Literal[
    "fixed_canonical",
    "fixed_0.1_0.9",
    "uniform_no_diag",
    "uniform_full",
    "location_weighted_uniform_no_diag",
    "location_weighted_uniform_full",
]
# Fixed canonical: Cycle down the list of receptacles
# Fixed 0.1 0.9: 90% to next, 10% stay
# Uniform no diag: Sample transition matrix from a uniform distribution, no self-transitions
# Uniform full: Sample transition matrix from a uniform distribution, self-transitions allowed
# Location weighted uniform no diag: Sample transition matrix from a uniform distribution weighted by location weights,
# location weighted uniform full: Sample transition matrix from a uniform distribution weighted by location weights, self-transitions allowed
MODE_DURATION_LITERALS = Literal["deterministic", "gaussian", "instant", "even"]
# This is for time:
# Deterministic: Always transition at the mean time
# Gaussian: Sample from a Gaussian with mean and std
# Instant: Always transition at the current timestamp
# Even: All modes have equal time
DT_HANDLING_STRATEGY = Literal["squash", "clip", "none"]

# Only add noise at the day level, make this more configurable
SCALE_TO_NOIS_STD = {
    "year": 0.0,     
    "month": 0.0,    
    "week": 0.0,     
    "day": 0.007,    
    "hour": 0.0,     
}

def normalize_vec(vec: jnp.ndarray) -> jnp.ndarray:
    return vec / vec.sum()


def weight_vec(vec: jnp.ndarray, w: jnp.ndarray) -> jnp.ndarray:
    return vec * w


def mk_normalized_vec_func(size):
    def gen_normalized_vec(key):
        key, rng = jax.random.split(key)
        weights = jax.random.uniform(rng, minval=0, maxval=1, shape=(size,))
        return normalize_vec(weights)

    return gen_normalized_vec


def generate_time_pattern(
    key,
    num_states: int,
    mode_duration_mode: MODE_DURATION_LITERALS,
    transition_model_mode: TRANSITION_MODEL_LITERALS,
    randomly_delete_transitions: Union[float, bool],
    locations_weights: Optional[jnp.ndarray],
    num_pad_locations: int = 0,
    scale: str = "hour"
):
    """
    Generate a normalized time pattern across a given number of states.
    The times are random but sum to 1.
    """
    key, rng = jax.random.split(key)
    weights = mk_normalized_vec_func(num_states)(rng)

    key, rngs = split_key(rng, num_states)
    transition_mat = jax.vmap(mk_normalized_vec_func(num_states))(rngs)
    if transition_model_mode in [
        "location_weighted_uniform_no_diag",
        "location_weighted_uniform_full",
    ]:  
        transition_mat = jax.vmap(lambda x, w: normalize_vec(weight_vec(x, w)))(
            jnp.ones_like(transition_mat), jnp.tile(locations_weights, num_states).reshape((num_states, num_states))
        )

    if randomly_delete_transitions:
        TO_DELETE = []
        for i, _ in enumerate(locations):
            if len(locations) - len(TO_DELETE) == 2:
                break
            key, rng = jax.random.split(key)
            if jax.random.uniform(rng, minval=0, maxval=1, shape=(1,)).item() > randomly_delete_transitions:
                TO_DELETE.append(i)
    else:
        TO_DELETE = []

    for to_delete in TO_DELETE:
        for l in range(num_states):
            transition_mat = transition_mat.at[l, to_delete].set(0)

    # Deterministic transition
    if transition_model_mode == "fixed_canonical":
        transition_mat = transition_mat * 0
        for l1 in range(num_states):
            l2 = (l1 + 1) % num_states
            while l2 in TO_DELETE:
                l2 = (l2 + 1) % num_states
            transition_mat = transition_mat.at[l1, l2].set(1)

    # Fixed probabilistic transition: 90% to next, 10% stay
    elif transition_model_mode == "fixed_0.1_0.9":
        transition_mat = transition_mat * 0
        for l1 in range(num_states):
            l2 = (l1 + 1) % num_states
            while l2 in TO_DELETE:
                l2 = (l2 + 1) % num_states

            transition_mat = transition_mat.at[l1, l1].set(0.1)
            transition_mat = transition_mat.at[l1, l2].set(0.9)

    # Uniform transition but no self-transitions
    elif transition_model_mode in ["uniform_no_diag", "location_weighted_uniform_no_diag"]:
        for l1 in range(num_states):
            transition_mat = transition_mat.at[l1, l1].set(0)
            transition_mat = jax.vmap(normalize_vec)(transition_mat)

    # Uniform transition including self-transitions
    elif transition_model_mode in ["uniform_full", "location_weighted_uniform_full"]:
        pass
    else:
        raise NotImplementedError(f"Unknown transition model mode: {transition_model_mode}")

    padded_shape = num_pad_locations + num_states
    # if padded_shape != num_states:
    weights = pad_to_shape(weights, (padded_shape,))
    transition_mat = pad_to_shape(transition_mat, (padded_shape, padded_shape))

    if mode_duration_mode == "gaussian":
        key, rng = jax.random.split(key)
        # FIXME: make stds an argument?
        stds = jnp.array([SCALE_TO_NOIS_STD[scale]] * num_states)
        # stds = jax.random.uniform(rng, minval=0, maxval=0.1, shape=(num_states,))
        ret = {"means": weights, "stds": pad_to_shape(stds, (padded_shape,)), "transition_mat": transition_mat}
    elif mode_duration_mode == "deterministic":
        ret = {"means": weights, "stds": jnp.zeros_like(weights), "transition_mat": transition_mat}
    elif mode_duration_mode == "instant":
        ret = {"means": jnp.ones_like(weights) * -10, "stds": jnp.zeros_like(weights), "transition_mat": transition_mat}
    elif mode_duration_mode == "even":
        ret = {
            "means": normalize_vec(jnp.ones_like(weights)),
            "stds": jnp.zeros_like(weights),
            "transition_mat": transition_mat,
        }
    else:
        raise NotImplementedError(f"mode_duration_mode {mode_duration_mode} not implemented")

    return ret

def handle_dt(scale, duration_and_transition_dict, dt, dt_handling_strategy, recursion_depth=0):
    MAX_RECURSION_DEPTH = 1000
    if recursion_depth > MAX_RECURSION_DEPTH:
        # give up
        return handle_dt(scale, duration_and_transition_dict, dt, "squash")

    if dt is None:
        if dt_handling_strategy != "none":
            print(f"WARNING: DT handling strategy was {dt_handling_strategy} but dt was not specified; dt was not handled.", file=sys.stderr)
        return duration_and_transition_dict

    from semistaticsim.groundtruth.schedule import TIME_SCALES_MAPPING
    DT_AT_CURRENT_SCALE = dt / TIME_SCALES_MAPPING[scale]
    EPS = 0.0001
    means, transition_mat = duration_and_transition_dict["means"], duration_and_transition_dict["transition_mat"]

    locs_under_the_dt = jnp.nonzero(jnp.logical_and(means <= DT_AT_CURRENT_SCALE, means > 0))[0]
    if len(locs_under_the_dt) == 0:
        return duration_and_transition_dict

    if dt_handling_strategy == "squash":
        for loc_index in locs_under_the_dt:
            means = means.at[loc_index].set(0)
            transition_mat = transition_mat.at[:, loc_index].set(0)
            transition_mat = transition_mat.at[loc_index, :].set(0)

        scc = get_scc(transition_mat)
        assert len(scc) >= 2, "The squash dt handling strategy inside schedule_generation.py has resulted in a transition matrix without any cycles. Consider using a transition model that results in more connections, or using the clip dt handling strategy."

        return {"means": means, "stds": duration_and_transition_dict["stds"], "transition_mat": transition_mat}

    elif dt_handling_strategy == "clip":
        # note: this CAN be done analytically, but results in completely unreadable code

        indices_that_were_0 = means <= 0

        distance_to_DT = DT_AT_CURRENT_SCALE - means[locs_under_the_dt] # compute bump to do
        means = (means - distance_to_DT.min() / means.shape[0]).clip(0,1) # remove even weight from all indices
        means = means.at[locs_under_the_dt].set(means[locs_under_the_dt] + distance_to_DT + EPS)
        means = means.at[indices_that_were_0].set(0) # should be useless but better be safe than sorry
        means = normalize_vec(means)

        assert jnp.all(indices_that_were_0 == (means <= 0))

        if recursion_depth > 0:
            print(f"DT Handling: recursion depth {recursion_depth} out of {MAX_RECURSION_DEPTH}")

        # by renormalizing, we might have pushed some values under EPS again. We can just iteratively do this process again and again until convergence.
        return handle_dt(scale, {"means": means, "stds": duration_and_transition_dict["stds"], "transition_mat": transition_mat}, dt, dt_handling_strategy, recursion_depth=recursion_depth+1)


def full_pattern(
    key,
    scales: List[str],
    locations: List[int],
    min_time_buckets: List[int],
    max_time_buckets: List[int],
    # leaf_order_mode: Literal["fixed_canonical", "uniform"] = "uniform",
    mode_duration_mode: MODE_DURATION_LITERALS = "deterministic",
    transition_model_mode: TRANSITION_MODEL_LITERALS = "fixed_canonical",
    # leaf_randomness_mode: Literal["uniform", "fixed_canonical"] = "uniform",
    randomly_delete_transitions: bool = False,  # This kills leaf - not transitions
    locations_weights: Optional[jnp.ndarray] = None,
    num_pad_locations: int = 0,
    dt_handling_strategy: DT_HANDLING_STRATEGY = "squash",
    dt: float = None
):
    """
    Recursively generate a hierarchical time pattern.

    Args:
        scales (list[str]): List of scales, from largest to smallest.
            Example: ["year", "month", "day"]
        locations (list[str]): Physical locations (only at the leaf scale).
        min_time_buckets (int): Minimum number of splits for higher scales.
        max_time_buckets (int): Maximum number of splits for higher scales.
        seed (int, optional): Random seed for reproducibility.
        is_gaussian (bool): Is the distribution gaussian or deterministic?.

    Returns:
        dict: Nested dictionary representing the full pattern.
    """
    # fixme need to make all DFS branches alter the same global key (maybe unroll? or just use a local global)
    # because otherwise the same key pattern occurs!
    # or actually maybe just the forloop: unnest and split before each downcall
    # if leaf_randomness_mode == "weighted_canonical":
    #    assert canonical_location_weights is not None
    #    if not isinstance(canonical_location_weights, jnp.ndarray):
    #        canonical_location_weights = jnp.array(canonical_location_weights)
    assert isinstance(locations[0], int) or isinstance(locations[0], float)
    assert len(scales) == len(min_time_buckets) == len(max_time_buckets)

    current_scale = scales[0]

    def ensure_cfgs_length(cfg_var):
        if not isinstance(cfg_var, list):
            cfg_var = [cfg_var] * len(scales)
        return cfg_var

    mode_duration_mode = ensure_cfgs_length(mode_duration_mode)
    transition_model_mode = ensure_cfgs_length(transition_model_mode)
    if locations_weights is None:
        locations_weights = jnp.ones((len(locations),))
    elif not isinstance(locations_weights, jnp.ndarray):
        locations_weights = jnp.array(locations_weights)

    # Leaf scale: assign location distribution
    if len(scales) == 1:

        key, rng = jax.random.split(key)
        distribution = generate_time_pattern(
            rng,
            len(locations),
            mode_duration_mode[0],
            transition_model_mode[0],
            randomly_delete_transitions,
            locations_weights=locations_weights,
            num_pad_locations=num_pad_locations,
            scale=current_scale,
        )

        distribution = handle_dt(scales[0], distribution, dt, dt_handling_strategy)
        # print("LOCATIONS AT LEAF", locations) fixme need to look at resolvable locs

        return {current_scale: {"distribution": distribution, "locations": locations}}

    # Higher scales: split into random buckets
    key, rng = jax.random.split(key)
    num_buckets = jax.random.randint(
        rng, minval=min_time_buckets[0], maxval=max_time_buckets[0], shape=(1,)
    ).item()  # random.randint(min_time_buckets, max_time_buckets)
    key, rng = jax.random.split(key)
    distribution = generate_time_pattern(
        rng,
        num_buckets,
        mode_duration_mode[0],
        transition_model_mode[0],
        randomly_delete_transitions,
        locations_weights=jnp.ones(num_buckets),
        num_pad_locations=0,
        scale=current_scale,
    )

    distribution = handle_dt(scales[0], distribution, dt, dt_handling_strategy)

    subpatterns = []
    for _ in range(num_buckets):
        key, rng = jax.random.split(key)
        subpat = full_pattern(
            rng,
            scales[1:],
            locations,
            min_time_buckets[1:],
            max_time_buckets[1:],
            mode_duration_mode[1:],
            transition_model_mode[1:],
            randomly_delete_transitions=randomly_delete_transitions,
            locations_weights=locations_weights,
            num_pad_locations=num_pad_locations,
            dt_handling_strategy=dt_handling_strategy,
            dt=dt
        )
        subpatterns.append(subpat)

    MISSING_BUCKETS = max_time_buckets[0] - num_buckets
    if MISSING_BUCKETS > 0:

        def pad_missing(v):
            if len(v.shape) == 1:
                return pad_to_shape(v, (max_time_buckets[0],))
            elif len(v.shape) == 2:
                return pad_to_shape(v, (max_time_buckets[0], max_time_buckets[0]))
            raise AssertionError()

        distribution = {k: pad_missing(v) for k, v in distribution.items()}
        for _ in range(MISSING_BUCKETS):  # won't ever get resolved to because of 0 transitions
            subpatterns.append(copy.deepcopy(subpatterns[-1]))

    return {current_scale: {"distribution": distribution, "subpatterns": subpatterns}}


def pad_to_shape(x: jnp.ndarray, target_shape: tuple[int, ...]) -> jnp.ndarray:
    """
    Pads `x` with zeros until it reaches `target_shape`.

    Args:
        x: jnp.ndarray, the input tensor
        target_shape: tuple of ints, desired output shape (must be >= x.shape)

    Returns:
        Padded tensor of shape `target_shape`.
    """
    if len(target_shape) != x.ndim:
        raise ValueError(f"target_shape {target_shape} must have same rank as x.shape {x.shape}")

    pad_widths = []
    for dim_size, target_size in zip(x.shape, target_shape):
        if target_size < dim_size:
            raise ValueError(f"target size {target_size} is smaller than input size {dim_size}")
        pad_before = 0
        pad_after = target_size - dim_size
        pad_widths.append((pad_before, pad_after))

    return jnp.pad(x, pad_widths, mode="constant", constant_values=0)


def flatten_pattern_for_visualization(pattern):
    """
    Flatten pattern into a format optimized for visualization.

    Returns:
        dict: Contains 'nodes', 'edges', and 'levels' for easy visualization
    """
    nodes = []
    edges = []
    levels = {}
    node_id = 0

    def _traverse_for_viz(node, path=None, parent_id=None, level=0):
        nonlocal node_id

        if path is None:
            path = []

        for scale, content in node.items():
            current_id = node_id
            node_id += 1

            current_path = path + [scale]
            distribution = content["distribution"]

            # Create node entry
            node_entry = {
                "id": current_id,
                "path": "/".join(current_path),
                "scale": scale,
                "level": level,
                "means": (
                    distribution["means"].tolist()
                    if hasattr(distribution["means"], "tolist")
                    else distribution["means"]
                ),
                "transition_matrix": (
                    distribution["transition_mat"].tolist()
                    if hasattr(distribution["transition_mat"], "tolist")
                    else distribution["transition_mat"]
                ),
                "parent_id": parent_id,
            }

            # Add locations if this is a leaf
            if "locations" in content:
                node_entry["locations"] = content["locations"]
                node_entry["is_leaf"] = True
            else:
                node_entry["is_leaf"] = False

            nodes.append(node_entry)

            # Track level information
            if level not in levels:
                levels[level] = []
            levels[level].append(current_id)

            # Create edge to parent if exists
            if parent_id is not None:
                edges.append({"from": parent_id, "to": current_id, "level_from": level - 1, "level_to": level})

            # Process subpatterns
            if "subpatterns" in content:
                for i, subpattern in enumerate(content["subpatterns"]):
                    bucket_path = current_path + [f"bucket_{i}"]
                    _traverse_for_viz(subpattern, bucket_path, current_id, level + 1)

    _traverse_for_viz(pattern)

    return {"nodes": nodes, "edges": edges, "levels": levels, "max_level": max(levels.keys()) if levels else 0}


def export_pattern_for_graphviz(pattern, filename="pattern.dot"):
    """
    Export pattern to Graphviz DOT format for visualization.
    """
    viz_data = flatten_pattern_for_visualization(pattern)

    with open(filename, "w") as f:
        f.write("digraph PatternTree {\n")
        f.write("  rankdir=TB;\n")
        f.write("  node [shape=box];\n\n")

        # Write nodes
        for node in viz_data["nodes"]:
            label = f"{node['scale']}"
            if node["is_leaf"]:
                label += f"\\nLocs: {node['locations']}"

            # Add some distribution info - only show first three means for brevity
            means_str = ", ".join([f"{x:.3f}" for x in node["means"][:3]])
            label += f"\\nMeans: [{means_str}...]"

            f.write(f'  {node["id"]} [label="{label}"];\n')

        f.write("\n")

        # Write edges
        for edge in viz_data["edges"]:
            f.write(f'  {edge["from"]} -> {edge["to"]};\n')

        f.write("}\n")

    print(f"Pattern exported to {filename}")


if __name__ == "__main__":
    # Example usage
    scales = ["month", "week", "day"]
    locations = [0, 1, 2]  # ["home", "work", "key rack"]
    # locations_weights = [0.1, 0.05, 0.2]

    pattern = full_pattern(
        key=jax.random.PRNGKey(0),
        scales=scales,
        locations=locations,
        min_time_buckets=[2, 5, 4],
        max_time_buckets=[2, 5, 4],
        # locations_weights=locations_weights,
        transition_model_mode="fixed_canonical",
    )

    # Export for visualization
    export_pattern_for_graphviz(pattern, "pattern_tree.dot")

    # Get visualization-optimized format
    viz_data = flatten_pattern_for_visualization(pattern)
    print("Visualization data structure:")
    print(f"Number of nodes: {len(viz_data['nodes'])}")
    print(f"Number of edges: {len(viz_data['edges'])}")
    print(f"Number of levels: {viz_data['max_level'] + 1}")
    print(f"Nodes per level: {viz_data['levels']}")



import threading
import time

from tqdm import tqdm

def spinner(desc, stop_event):
    from itertools import cycle

    spinner = cycle(['⠋', '⠙', '⠹', '⠸', '⠼', '⠴', '⠦', '⠧', '⠇', '⠏'])
    with tqdm(total=None, desc=desc, bar_format='{desc}') as pbar:
        while not stop_event.is_set():
            pbar.set_description(f"{desc} - {next(spinner)}")
            time.sleep(0.1)
            #pbar.update()

def start_spinner_thread(desc) -> (threading.Event, threading.Thread):
    stop_event = threading.Event()
    spinner_thread = threading.Thread(
        target=spinner,
        args=(desc, stop_event)
    )
    spinner_thread.start()

    return stop_event, spinner_thread

def stop_spinner_thread(stop_event, spinner_thread=None):
    if isinstance(stop_event, tuple):
        stop_event, spinner_thread = stop_event
    stop_event.set()
    spinner_thread.join(timeout=5)


import copy
import functools
from functools import partial
from typing import List, Tuple, Set

import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import shapely
from flax import struct
from shapely.geometry import Polygon
from shapely.ops import triangulate
from typing_extensions import Self

from semistaticsim.rendering.utils.dict2xyztup import dict2xyztuple


@partial(jax.jit, static_argnums=(1,))
def split_key(key, num_keys):
    key, *rng = jax.random.split(key, num_keys + 1)
    rng = jnp.reshape(jnp.stack(rng), (num_keys, 2))
    return key, rng

def point_to_shapely(point):
    if len(point) == 3:
        def cast(p):
            return shapely.Point(p[0], p[2])
    else:
        def cast(p):
            return shapely.Point(p[0], p[1])
    return cast(point)


def pad_to_shape(x: jnp.ndarray, target_shape: tuple[int, ...]) -> jnp.ndarray:
    """
    Pads `x` with zeros until it reaches `target_shape`.

    Args:
        x: jnp.ndarray, the input tensor
        target_shape: tuple of ints, desired output shape (must be >= x.shape)

    Returns:
        Padded tensor of shape `target_shape`.
    """
    if len(target_shape) != x.ndim:
        raise ValueError(f"target_shape {target_shape} must have same rank as x.shape {x.shape}")

    pad_widths = []
    for dim_size, target_size in zip(x.shape, target_shape):
        if target_size < dim_size:
            raise ValueError(f"target size {target_size} is smaller than input size {dim_size}")
        pad_before = 0
        pad_after = target_size - dim_size
        pad_widths.append((pad_before, pad_after))

    return jnp.pad(x, pad_widths, mode="constant", constant_values=0)



@struct.dataclass
class RoomPolygon:
    scene_room_id: int
    scene_room_floorPolygon: list[dict]
    triangles: jnp.ndarray
    triangle_areas: jnp.ndarray

    @classmethod
    def create(cls, id, room):
        self = RoomPolygon(id, room, None, None)

        polygon = Polygon(self.floor_polygon_coords)
        def polygon_to_triangles(polygon):
            """
            Triangulate a shapely polygon and return list of triangles.
            """
            tris = shapely.ops.triangulate(polygon)

            ret = []
            for tri in tris:
                centroid = tri.centroid
                if not polygon.contains(centroid):
                    continue
                ret.append(list(tri.exterior.coords)[:-1])

            return ret
        tris = jnp.array(polygon_to_triangles(polygon))

        def triangle_area(tri):
            x1, y1 = tri[0]
            x2, y2 = tri[1]
            x3, y3 = tri[2]
            return 0.5 * jnp.abs((x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1))
        triangle_areas = jax.vmap(triangle_area)(tris)

        return self.replace(triangles=tris, triangle_areas=triangle_areas)

    @property
    def num_triangles(self):
        return self.triangles.shape[0]

    @property
    def room_bbox(self):
        coords = self.floor_polygon_coords

        min_x, min_y = jnp.inf, jnp.inf
        max_x, max_y = -jnp.inf, -jnp.inf

        for coord in coords:
            min_x = min(min_x, coord[0])
            min_y = min(min_y, coord[1])
            max_x = max(max_x, coord[0])
            max_y = max(max_y, coord[1])

        return min_x, min_y, max_x, max_y

    def containment_mask(self, points):
        return [self.shapely_floor_polygon.contains(point_to_shapely(point)) for point in points]

    def _points_distance_to_polygon_edge(self, points):
        polygon = self.shapely_floor_polygon
        return [point_to_shapely(point).distance(polygon.boundary) for point in points]

    def points_distance_to_polygon_edge_mask(self, points, mindist):
        dists = self._points_distance_to_polygon_edge(points)
        return [d > mindist for d in dists]

    #@property
    def networkx(self, grid_size=0.05):
        min_x, min_y, max_x, max_y = self.room_bbox

        x = np.arange(min_x + grid_size * 2, max_x - grid_size, grid_size)
        y = np.arange(min_y + grid_size * 2, max_y - grid_size, grid_size)

        DEFAULT_ROBOT_HEIGHT = 0.95

        # Create 2D grid
        xx, yy = np.meshgrid(x, y)
        grid_points = np.stack([xx.ravel(), yy.ravel()], axis=-1, dtype=np.float32)
        grid_points = [(x, DEFAULT_ROBOT_HEIGHT, z) for x, z in grid_points]

        mask = self.containment_mask(grid_points)
        grid_points = [p for i, p in enumerate(grid_points) if mask[i]]
        mask = self.points_distance_to_polygon_edge_mask(grid_points, 0.3)
        grid_points = [p for i, p in enumerate(grid_points) if mask[i]]
        grid_points = [(z,y,x) for (x,y,z) in grid_points]

        G = nx.Graph()
        points_set = set(grid_points)

        # Add all points as nodes
        for p in points_set:
            G.add_node(p)

        # Define neighbor directions (axis-aligned)
        DIM = 3
        directions = []
        for d in range(DIM):
            if d == 1:
                continue
            for offset in [-grid_size, grid_size]:
                dir_vector = [0] * DIM
                dir_vector[d] = offset
                directions.append(tuple(dir_vector))

        # Add edges between neighbors
        for p in points_set:
            for d in directions:
                neighbor = tuple(np.add(p, d))
                if neighbor in points_set:
                    G.add_edge(p, neighbor)

        return G

    @property
    def floor_polygon_coords(self):
        return list(map(lambda x: (x["x"], x["z"]), self.scene_room_floorPolygon))

    @property
    def shapely_floor_polygon(self):
        return Polygon(self.floor_polygon_coords)

    @property
    def shapely_triangles(self):
        ret = []
        for tri in self.triangles:
            ret.append(Polygon(tri))
        return ret

    def plot_self(self, samples=None, plot_triangles=False, reachability_graph=None, show=False):
        plt.plot(*self.shapely_floor_polygon.exterior.xy)
        if plot_triangles:
            for tri in self.shapely_triangles:
                x, y = tri.exterior.xy
                plt.fill(x, y, alpha=0.2)

        if reachability_graph is not None:

            def draw_grid_graph_2d(G, path=None, node_size=100, node_color='lightblue', edge_color='gray',
                                   show_labels=False):
                pos = {node: (node[2], node[0]) for node in G.nodes}  # node positions are their coordinates

                nx.draw(G, pos,
                        node_size=node_size,
                        node_color=node_color,
                        edge_color=edge_color,
                        with_labels=show_labels,
                        font_size=8)

                # If a path is provided, draw it in red
                if path:
                    # Draw path edges first (so nodes will be on top)
                    path_edges = list(zip(path[:-1], path[1:]))
                    nx.draw_networkx_edges(G, pos,
                                           edgelist=path_edges,
                                           edge_color='red',
                                           width=2)
                    # Draw path nodes
                    nx.draw_networkx_nodes(G, pos,
                                           nodelist=path,
                                           node_size=node_size,
                                           node_color='red')

            draw_grid_graph_2d(reachability_graph)

        if samples is not None:
            plt.scatter(samples[:,0], samples[:,1], c="orange", zorder=10)

        if show:
            plt.show()

    @functools.partial(jax.jit, static_argnums=2)
    def sample_from_room(self, key, num_samples=1):
        if num_samples > 1:
            _, keys = split_key(key, num_samples)
            return jax.vmap(self.sample_from_room)(keys).squeeze()

        probs = self.triangle_areas / self.triangle_areas.sum()

        key_tri, key_uv = jax.random.split(key)
        tri_indices = jax.random.choice(key_tri, self.triangles.shape[0], (1,), p=probs)
        chosen_tris = self.triangles[tri_indices]

        uvs = jax.random.uniform(key_uv, (1, 2))
        u = jnp.sqrt(uvs[:, 0])
        v = uvs[:, 1]
        w0 = 1 - u
        w1 = u * (1 - v)
        w2 = u * v
        weights = jnp.stack([w0, w1, w2], axis=-1)
        weights_expanded = weights[:, :, jnp.newaxis]

        sampled_points = jnp.sum(chosen_tris * weights_expanded, axis=1)
        return sampled_points  # Returns [1, 2] array

@struct.dataclass
class DoorWay:
    holePolygon: List[Tuple[float,float,float]]
    position_center: Tuple[float,float]
    waypoints: List[Tuple[float,float]]

    def __contains__(self, point_node):
        if len(point_node) == 3:
            point_node = (point_node[0], point_node[2])
        return point_node in self.waypoints

    @classmethod
    def create(cls, door, grid_size):
        holePolygon = door["holePolygon"]
        p1, p2 = holePolygon

        center = ((p1['x'] + p2['x']) / 2, (p1['z'] + p2['z']) / 2)
        center = (round(center[0],1), round(center[1],1))

        # doors are always axis-aligned
        if p1['x'] == p2['x']:
            waypoints = [
                (center[0], center[1] + grid_size),
                center,
                (center[0], center[1] - grid_size),
            ]
        elif p1['z'] == p2['z']:
            waypoints = [
                (center[0] + grid_size, center[1]),
                center,
                (center[0] - grid_size, center[1]),
            ]
        else:
            raise AssertionError("Non-axis aligned door detected, failing.")

        return DoorWay(holePolygon, center, waypoints)

@struct.dataclass
class FloorPolygon:
    room_polygons: List[RoomPolygon]
    #doorways: List[DoorWay]
    grid_size: float

    @property
    def room_areas(self):
        return jnp.array([jnp.sum(room.triangle_areas) for room in self.room_polygons])

    @property
    def room_polygon_coords(self):
        return [room.floor_polygon_coords for room in self.room_polygons]

    @property
    def doorway_waypoints(self):
        return [x.waypoints for x in self.doorways]

    @property
    def flat_room_polygon_coords(self):
        ret = []
        for room in self.room_polygon_coords:
            ret += room
        return ret

    @property
    def floor_bbox(self):
        coords = self.flat_room_polygon_coords

        min_x, min_y = jnp.inf, jnp.inf
        max_x, max_y = -jnp.inf, -jnp.inf

        for coord in coords:
            min_x = min(min_x, coord[0])
            min_y = min(min_y, coord[1])
            max_x = max(max_x, coord[0])
            max_y = max(max_y, coord[1])

        return min_x, min_y, max_x, max_y

    @property
    def triangles(self):
        return jnp.concatenate([room.triangles for room in self.room_polygons])

    @property
    def triangle_areas(self):
        return jnp.concatenate([room.triangle_areas for room in self.room_polygons])

    @functools.partial(jax.jit, static_argnums=2)
    def sample_from_floor(self, key, num_samples=1, triangles=None, triangle_areas=None):
        if triangles is None:
            triangles = self.triangles
        if triangle_areas is None:
            triangle_areas = self.triangle_areas

        # what is a collecton of rooms, but a big room? much to ponder.
        return self.room_polygons[0].replace(triangles=triangles, triangle_areas=triangle_areas).sample_from_room(key=key, num_samples=num_samples)

    @classmethod
    def create(cls, scene, grid_size=0.1):
        rooms = scene["rooms"]

        MAX_NUM_TRIANGLES = -jnp.inf
        selves = []
        for i, room in enumerate(rooms):
            self = RoomPolygon.create(i, room["floorPolygon"])
            MAX_NUM_TRIANGLES = max(MAX_NUM_TRIANGLES, self.num_triangles)
            selves.append(self)

        for i in range(len(selves)):
            self = selves[i]
            self = self.replace(triangles=pad_to_shape(self.triangles, (MAX_NUM_TRIANGLES, 3, 2)))
            self = self.replace(triangle_areas=pad_to_shape(self.triangle_areas, (MAX_NUM_TRIANGLES,)))
            selves[i] = self

        doorways = [DoorWay.create(door, grid_size) for door in scene["doors"]]

        return FloorPolygon(selves, grid_size)

    def containment_mask(self, points):
        room_masks = jnp.array([room.containment_mask(points) for room in self.room_polygons])
        floor_mask = jnp.sum(room_masks, axis=0) > 0
        assert floor_mask.shape[0] == len(points)
        return floor_mask

    def points_distance_to_polygon_edge_mask(self, points, mindist):
        room_masks = jnp.array([room.points_distance_to_polygon_edge_mask(points, mindist) for room in self.room_polygons])
        floor_mask = jnp.sum(room_masks, axis=0) == room_masks.shape[0]
        assert floor_mask.shape[0] == len(points)
        return floor_mask

    #@functools.lru_cache(maxsize=None)
    def rechability_graph(self, mindist_to_walls=0.4, DEFAULT_ROBOT_HEIGHT=0.95, set_mindist_to_obj=0.3, pre_selected_points=None):
        grid_size = self.grid_size

        if pre_selected_points is None:
            min_x, min_y, max_x, max_y = self.floor_bbox

            min_x = round(min_x - 1)
            min_y = round(min_y - 1)
            max_x = round(max_x + 1)
            max_y = round(max_y + 1)

            x = np.arange(min_x + grid_size * 2, max_x - grid_size, grid_size)
            y = np.arange(min_y + grid_size * 2, max_y - grid_size, grid_size)

            # Create 2D grid
            xx, yy = np.meshgrid(x, y)
            grid_points = np.stack([xx.ravel(), yy.ravel()], axis=-1, dtype=np.float32)
            grid_points = [(x, DEFAULT_ROBOT_HEIGHT, z) for x, z in grid_points]

            mask = self.containment_mask(grid_points)
            grid_points = [p for i, p in enumerate(grid_points) if mask[i]]
            mask = self.points_distance_to_polygon_edge_mask(grid_points, mindist_to_walls)
            grid_points = [p for i, p in enumerate(grid_points) if mask[i]]
            grid_points = [(float(z),float(y),float(x)) for (x,y,z) in grid_points]
            grid_points = [(z,y,x) for (x,y,z) in grid_points]
        else:
            assert isinstance(pre_selected_points, list) and isinstance(pre_selected_points[0], tuple)
            grid_points = pre_selected_points


        G = nx.Graph()
        points_set = set(grid_points)

        # Add all points as nodes
        for p in points_set:
            G.add_node(p)

        # Define neighbor directions (axis-aligned)
        DIM = 3
        directions = []
        for d in range(DIM):
            if d == 1:
                continue
            for offset in [-grid_size, grid_size]:
                dir_vector = [0] * DIM
                dir_vector[d] = offset
                directions.append(tuple(dir_vector))

        # Add edges between neighbors
        for p in points_set:
            for d in directions:
                neighbor = tuple(np.add(p, d))
                if neighbor in points_set:
                    G.add_edge(p, neighbor)

        ALL_OTHER_NODES = list(G.nodes())
        JNP_ALL_OTHER_NODES = jnp.array(ALL_OTHER_NODES)

        @jax.jit
        def dist(target, other):
            return jnp.linalg.norm(target - other)

        def add_default_robot_height(p):
            return (p[0], DEFAULT_ROBOT_HEIGHT, p[1])

        # doorways time
        if False:
            for (out1, center, out2) in self.doorway_waypoints:
                out1 = add_default_robot_height(out1)
                center = add_default_robot_height(center)
                out2 = add_default_robot_height(out2)

                G.add_node(out1)
                G.add_node(center)
                G.add_node(out2)

                G.add_edge(out1, center)
                G.add_edge(out2, center)

                dist_to_out1 = jax.vmap(functools.partial(dist, jnp.array(out1)))(JNP_ALL_OTHER_NODES)
                dist_to_out2 = jax.vmap(functools.partial(dist, jnp.array(out2)))(JNP_ALL_OTHER_NODES)

                closest_node_index_out1 = jnp.argmin(dist_to_out1)
                closest_node_index_out2 = jnp.argmin(dist_to_out2)

                closest_node_out1 = ALL_OTHER_NODES[closest_node_index_out1]
                closest_node_out2 = ALL_OTHER_NODES[closest_node_index_out2]

                G.add_edge(out1, closest_node_out1)
                G.add_edge(out2, closest_node_out2)

        return ReachabilityGraph(G, grid_size, points_set, jnp.array(list(points_set)), set_mindist_to_obj, self)

    def plot_self(self, samples=None, plot_triangles=True, reachability_graph=None, show=True):
        """
        Plot the entire scene composed of multiple RoomPolygon instances.
        """
        # Plot each room
        for room in self.room_polygons:  # assuming self.rooms is a list of RoomPolygon
            room.plot_self(
                samples=None,  # per-room samples are not typically meaningful
                plot_triangles=plot_triangles,
                reachability_graph=None,  # or you could pass per-room graphs if available
                show=False,  # don't show until the end
            )

        # Optionally overlay scene-level data
        if reachability_graph is not None:
            def draw_grid_graph_2d(G, path=None, node_size=100, node_color='lightblue', edge_color='gray',
                                   show_labels=False):
                #pos = {node: (node[2], node[0]) for node in G.nodes}
                pos = {node: (node[0], node[2]) for node in G.nodes()}
                nx.draw(G, pos,
                        node_size=node_size,
                        node_color=node_color,
                        edge_color=edge_color,
                        with_labels=show_labels,
                        font_size=8)
                if path:
                    path_edges = list(zip(path[:-1], path[1:]))
                    nx.draw_networkx_edges(G, pos,
                                           edgelist=path_edges,
                                           edge_color='red',
                                           width=2)
                    nx.draw_networkx_nodes(G, pos,
                                           nodelist=path,
                                           node_size=node_size,
                                           node_color='red')
            if isinstance(reachability_graph, ReachabilityGraph):
                reachability_graph = reachability_graph.full_graph
            draw_grid_graph_2d(reachability_graph)

        if samples is not None:
            if not hasattr(samples, "shape"):
                samples = np.atleast_2d(np.asarray(samples))
            if samples.shape[1] == 3:
                x = samples[:, 0]
                y = samples[:, 2]
            else:
                x = samples[:, 0]
                y = samples[:, 1]
            plt.scatter(x,y, c="orange", zorder=10)

        if show:
            plt.show()


@struct.dataclass
class ReachabilityGraph:
    full_graph: nx.Graph
    grid_size: float
    point_set: Set[Tuple[float, float, float]]
    points: jnp.ndarray
    set_mindist_to_obj: float
    parent_floor: FloorPolygon

    @property
    def doorways(self) -> List[DoorWay]:
        return self.parent_floor.doorways

    def doorway_nodes(self):
        return [x.position for x in self.doorways]

    def shunt(self, point):
        assert len(point) == 3

        def dist(p1, p2):
            return jnp.linalg.norm(p1 - p2)

        in_room = None
        # Determine the room polygon the point is at
        for room in self.parent_floor.room_polygons:
            if room.shapely_floor_polygon.contains(point_to_shapely(point)):
                in_room = room
                break

        if in_room is None:
            print("I'm sorry, point is outside all known rooms.")
            raise AssertionError()
            # fallback: use all nodes if point is outside known rooms
            candidate_nodes = list(self.full_graph.nodes)
        else:
            # Filter graph nodes to only those within the same room
            # fixme this could be precomputed
            candidate_nodes = [
                n for n in self.full_graph.nodes
                if in_room.shapely_floor_polygon.contains(point_to_shapely(n))
            ]
            # If there are no nodes in that room, fallback to all nodes
            if len(candidate_nodes) == 0:
                candidate_nodes = list(self.full_graph.nodes)

        all_nodes = jnp.array(candidate_nodes)
        # all_nodes = jnp.array(list(self.full_graph.nodes))

        def shunt_fn(pos, universal_y):
            pos = jnp.array(pos)
            pos = pos.at[1].set(universal_y)

            dists = jnp.nan_to_num(jax.vmap(functools.partial(dist, p2=pos))(all_nodes), neginf=jnp.inf, nan=jnp.inf,
                                   posinf=jnp.inf)
            return dists.argmin()
        argret = shunt_fn(point, universal_y=list(self.nodes)[0][1])
        return candidate_nodes[int(argret)]

    @property
    def nodes(self):
        return self.full_graph.nodes

    def clean_convert(self, point):
        # fixes conversion errors between np, jnp, lists
        import jax.numpy as jnp
        if isinstance(point, jnp.ndarray):
            point = np.array(point)
        if isinstance(point, np.ndarray):
            point = point.tolist()
        rounded = [round(x / self.grid_size) * self.grid_size for x in point]
        return tuple(rounded)

    def filter_(self, objects_metadata, mindist_to_obj=None) -> Self:
        if mindist_to_obj is None:
            mindist_to_obj = self.set_mindist_to_obj

        def is_inside_aabb(obj_pos, obj_size, agent_pos):
            """
            Check if an agent at `agent_pos` is inside the AABB defined by `obj_pos` and `obj_size`.
            Returns True if inside, False if outside.
            """

            half_size = obj_size / 2
            x_inside = (obj_pos[0] - half_size[0] <= agent_pos[0]) & (agent_pos[0] <= obj_pos[0] + half_size[0])
            z_inside = (obj_pos[2] - half_size[2] <= agent_pos[2]) & (agent_pos[2] <= obj_pos[2] + half_size[2])

            return jnp.logical_and(x_inside, z_inside)

        @jax.jit
        def _filter_agent_positions(agent_positions, object_positions, object_sizes, margin=0.0):
            """
            Remove agent positions that are inside any object's AABB, considering a margin.
            """
            inflated_sizes = object_sizes + margin

            def is_inside_any_object(single_agent_pos, all_object_poses, all_object_sizes):
                return jax.vmap(functools.partial(is_inside_aabb, agent_pos=single_agent_pos))(all_object_poses,
                                                                                               all_object_sizes).any()

            mask = jax.vmap(functools.partial(is_inside_any_object, all_object_poses=object_positions,
                                              all_object_sizes=inflated_sizes))(agent_positions)
            return jnp.logical_not(mask)

        reachable_positions = copy.deepcopy(self.full_graph)
        todo_positions = reachable_positions.nodes

        DO_NOT_CONSIDER = lambda x: any(y in x.lower() for y in ["door", "painting", "window", "wall", "room", "floor"])

        centers = [dict2xyztuple(o["position"]) for o in objects_metadata if not DO_NOT_CONSIDER(o["objectId"])]
        sizes = [dict2xyztuple(o["axisAlignedBoundingBox"]["size"]) for o in objects_metadata if not DO_NOT_CONSIDER(o["objectId"])]
        #centers = [obj.position for name, obj in runtime_container.objects_map.items() if not DO_NOT_CONSIDER(name)]
        #sizes = [obj.size for name, obj in runtime_container.objects_map.items() if not DO_NOT_CONSIDER(name)]

        todo_positions = jnp.array(todo_positions)
        centers = jnp.array(centers)
        sizes = jnp.array(sizes)

        mask = _filter_agent_positions(todo_positions.at[:, 1].set(0.95), centers.at[:, 1].set(0.95), sizes,
                                       margin=mindist_to_obj) # fixme maybe 0.4 should be an arg?

        print("Num positions to remove", jnp.sum(jnp.logical_not(mask)))

        mask = np.array(mask)
        todo_positions = np.array(todo_positions)

        def removing_node_breaks_connectivity(graph, node_to_remove):
            graph = copy.deepcopy(graph)
            NUM_COMPONENTS = len(list(nx.connected_components(graph)))

            graph.remove_node(node_to_remove)
            NEW_NUM_COMPONENTS = len(list(nx.connected_components(graph)))

            return NUM_COMPONENTS != NEW_NUM_COMPONENTS

        for position in todo_positions[jnp.logical_not(mask)]:
            for node in nx.nodes(reachable_positions):
                if np.all(np.allclose(node, position)):
                    if not removing_node_breaks_connectivity(reachable_positions, node):
                        reachable_positions.remove_node(node)
                        break

        # Step 2: Keep only the largest connected component
        if reachable_positions.number_of_nodes() > 0:
            largest_cc_nodes = max(nx.connected_components(reachable_positions), key=len)
            reachable_positions = reachable_positions.subgraph(largest_cc_nodes).copy()

        return self.replace(full_graph=reachable_positions)



import cv2
import shutil
import math
from typing import Sequence
import numpy as np
from copy import deepcopy
import os
import hydra
import json
from omegaconf import DictConfig
import yaml
from pathlib import Path
from natsort import natsorted

from semistaticsim.groundtruth.schedule import TIME_SCALES_MAPPING
from semistaticsim.keyboardcontrol.postprocessing import format_results, purge_dir
from semistaticsim.rendering.ai2thor_hippo_controller import get_sim, random_teleport

actionList = {
    "MoveAhead": "w",
    "MoveBack": "s",
    "MoveLeft": "a",
    "MoveRight": "d",
    "RotateLeft": "q",
    "RotateRight": "e",
    "LookUp": "r",
    "LookDown": "f",
    "FINISH": "p",
    # Interact action
    "PickupObject": "z",
    "PutObject": "x",
    "OpenObject": "c",
    "CloseObject": "v",
    "ToggleObjectOn": "b",
    "ToggleObjectOff": "n",
    "SliceObject": "m",
    "GoToObject": "k",
    "StepSemiStaticObjects": "i",
    "GoToAllInstances": "l",
}

ROBOTS = [
    {
        "name": "robot1",
        "skills": [
            "GoToObject",
            "OpenObject",
            "CloseObject",
            "BreakObject",
            "SliceObject",
            "SwitchOn",
            "SwitchOff",
            "PickupObject",
            "PutObject",
            "DropHandObject",
            "ThrowObject",
            "PushObject",
            "PullObject",
        ],
    },
    {
        "name": "robot2",
        "skills": [
            "GoToObject",
            "OpenObject",
            "CloseObject",
            "BreakObject",
            "SliceObject",
            "SwitchOn",
            "SwitchOff",
            "PickupObject",
            "PutObject",
            "DropHandObject",
            "ThrowObject",
            "PushObject",
            "PullObject",
        ],
    },
]


def save_sim_results(
    sim, res_path: str, batch_size: int, cleanup_previous_data=False, generate_video=False, generate_gif=False
):
    # Delete ALL data including formatted runs
    if cleanup_previous_data and os.path.exists(res_path):
        shutil.rmtree(res_path)
    # Delete data but leave run folders
    elif os.path.exists(res_path):
        purge_dir(res_path, prefix="run_")
    os.makedirs(res_path, exist_ok=True)

    # Save id_to_color mapping if exists
    with open(os.path.join(sim.tmpdir, "id_to_color.json"), "w") as f:
        json.dump(sim.id_to_color, f, indent=4)

    if sim.privileged_apn and sim.privileged_apn.global_length > 0:
        sim.privileged_apn.dump_to_parquet(sim.tmpdir, dump_leftover=True, batch_size=batch_size)
    # Delete privileged_apn to save memory
    del sim.privileged_apn
    # Privileged data (presence/absence)
    format_results(res_path, sim.tmpdir, batch_size)

    # Videos & Gifs
    if generate_video or generate_gif:
        # Generate videos and gifs for each run
        run_folders = natsorted(Path(res_path).glob("run_*"))
        for run_folder in run_folders:
            vid_root = run_folder / "video"
            gif_root = run_folder / "gif"
            os.makedirs(vid_root, exist_ok=True)
            os.makedirs(gif_root, exist_ok=True)
            img_root = run_folder / "images"
            # Read the images and compile them into videos
            for item in os.listdir(img_root):
                frame_list = []
                img_dir = img_root / item
                img_files = sorted(os.listdir(img_dir))
                for img_file in img_files:
                    img_path = img_dir / img_file
                    frame = (
                        cv2.imread(str(img_path), cv2.IMREAD_UNCHANGED)[..., None]
                        if "depth" in item
                        else cv2.cvtColor(cv2.imread(str(img_path)), cv2.COLOR_RGB2BGR)
                    )
                    frame_list.append(frame)
                if generate_video:
                    export_video(str(vid_root / f"{item}.mp4"), frame_list, depth=("depth" in item))
                if generate_gif:
                    clip = show_video(frame_list, fps=5, depth=("depth" in item))
                    clip.write_gif(str(gif_root / f"{item}.gif"))


def get_interact_object(env, visible_only=True):
    candidates = []
    objectId = ""

    for obj in env.last_event.metadata["objects"]:
        if visible_only:
            if obj["visible"]:
                candidates.append(obj["objectId"])
        else:
            candidates.append(obj["objectId"])

    if len(candidates) == 0:
        print("no valid interact object candidates")
        return None
    else:
        print("===========choose index from the candidates==========")

        # Create mapping from characters to indices
        char_to_index = {}
        index = 0

        # Add digits 0-9
        for i in range(10):
            if index < len(candidates):
                char_to_index[str(i)] = index
                print(f"{i} : {candidates[index]}")
                index += 1

        # Add lowercase letters a-z (will also handle uppercase via normalization)
        for c in "abcdefghijklmnopqrstuvwxyz":
            if index < len(candidates):
                char_to_index[c] = index
                char_to_index[c.upper()] = index  # Map both cases to same index
                print(f"{c}/{c.upper()} : {candidates[index]}")
                index += 1

        # Add remaining uppercase letters if we have more than 36 objects
        for c in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
            if index < len(candidates) and c not in char_to_index:
                char_to_index[c] = index
                print(f"{c} : {candidates[index]}")
                index += 1

        # Define shift key codes (common shift key codes)
        SHIFT_CODES = {225, 65505, 65506}  # Left Shift, Right Shift, etc.

        while True:
            keystroke = cv2.waitKey()
            print(f"Raw keystroke: {keystroke}")

            if keystroke == actionList["FINISH"]:
                print("stop interact")
                break

            # Ignore shift keys
            if keystroke in SHIFT_CODES:
                print("Ignoring Shift key...")
                continue

            # Convert keystroke to character
            try:
                # Convert ASCII code to character
                if 0 <= keystroke <= 255:
                    char = chr(keystroke)
                    print(f"Converted to char: '{char}'")

                    # Normalize to lowercase for letters (since we map both cases)
                    if char.isalpha():
                        normalized_char = char.lower()
                    else:
                        normalized_char = char

                    # Check if character is in our mapping
                    if normalized_char in char_to_index:
                        object_index = char_to_index[normalized_char]
                        objectId = candidates[object_index]
                        print(f"Selected object: {objectId}")
                        break
                    else:
                        print(f"INVALID KEY '{char}' - not in available options")
                else:
                    print(f"INVALID KEYCODE {keystroke}")
            except Exception as e:
                print(f"ERROR processing keystroke: {e}")
                continue

        return objectId


def keyboard_play(
    env,
    sim=None,
):
    sim.kill_sim_on_condition_failure = False
    sim.raise_exception_on_condition_failure = False
    env.humanviewing.incr_action_idx()

    if sim is None:
        raise NotImplementedError()

    step = 0
    while True:
        keystroke = cv2.waitKey(0)
        step += 1

        if keystroke == ord(actionList["FINISH"]):
            env.stop()
            cv2.destroyAllWindows()
            print("action: STOP")
            break

        if keystroke == ord(actionList["MoveAhead"]):
            action = "MoveAhead"
            print("action: MoveAhead")
            sim.MoveAhead(ROBOTS[0], None)
        elif keystroke == ord(actionList["MoveBack"]):
            action = "MoveBack"
            print("action: MoveBack")
            sim.MoveBack(ROBOTS[0], None)
        elif keystroke == ord(actionList["MoveLeft"]):
            action = "MoveLeft"
            print("action: MoveLeft")
            sim.MoveLeft(ROBOTS[0], None)
        elif keystroke == ord(actionList["MoveRight"]):
            action = "MoveRight"
            print("action: MoveRight")
            sim.MoveRight(ROBOTS[0], None)
        elif keystroke == ord(actionList["RotateLeft"]):
            action = "RotateLeft"
            print("action: RotateLeft")
            sim.RotateLeft(ROBOTS[0], None)
        elif keystroke == ord(actionList["RotateRight"]):
            action = "RotateRight"
            print("action: RotateRight")
            sim.RotateRight(ROBOTS[0], None)
        elif keystroke == ord(actionList["LookUp"]):
            action = "LookUp"
            print("action: LookUp")
            sim.LookUp(ROBOTS[0])
        elif keystroke == ord(actionList["LookDown"]):
            action = "LookDown"
            print("action: LookDown")
            sim.LookDown(ROBOTS[0])

        elif keystroke == ord(actionList["GoToObject"]):
            action = "GoToObject"
            print("Go to which object?")
            objectId = get_interact_object(env, visible_only=False)
            sim.GoToObject(ROBOTS[0], objectId)
        elif keystroke == ord(actionList["PickupObject"]):
            action = "PickupObject"
            print("action: PickupObject")
            objectId = get_interact_object(env)
            sim.PickupObject(ROBOTS[0], objectId)
            # if len(sim.exception_queue) <= 0:
            pickedupobj = objectId
        elif keystroke == ord(actionList["PutObject"]):
            action = "PutObject"
            print("action: PutObject")
            objectId = get_interact_object(env)
            # if len(sim.exception_queue) <= 0:
            sim.PutObject(ROBOTS[0], pickedupobj, objectId)
            pickedupobj = None
        elif keystroke == ord(actionList["OpenObject"]):
            action = "OpenObject"
            print("action: OpenObject")
            objectId = get_interact_object(env)
            sim.OpenObject(ROBOTS[0], objectId)
        elif keystroke == ord(actionList["CloseObject"]):
            action = "CloseObject"
            print("action: CloseObject")
            objectId = get_interact_object(env)
            sim.CloseObject(ROBOTS[0], objectId)
        elif keystroke == ord(actionList["ToggleObjectOn"]):
            action = "ToggleObjectOn"
            print("action: ToggleObjectOn")
            objectId = get_interact_object(env)
            sim.ToggleObjectOn(ROBOTS[0], objectId)
        elif keystroke == ord(actionList["ToggleObjectOff"]):
            action = "ToggleObjectOff"
            print("action: ToggleObjectOff")
            objectId = get_interact_object(env)
            sim.ToggleObjectOff(ROBOTS[0], objectId)
        elif keystroke == ord(actionList["StepSemiStaticObjects"]):
            action = "StepSemiStaticObjects"
            print("action: StepSemiStaticObjects")
            sim.StepSemiStaticObjects()
        elif keystroke == ord(actionList["SliceObject"]):
            action = "SliceObject"
            print("action: SliceObject")
            objectId = get_interact_object(env)
            sim.SliceObject(ROBOTS[0], objectId)
        elif keystroke == ord(actionList["GoToAllInstances"]):
            action = "GoToAllInstances"
            print("action: GoToAllInstances")
            sim.GoToAllInstances(ROBOTS[0])
        else:
            print("INVALID KEY", keystroke)
            continue

        print("Interact object:")
        try:
            print(objectId)
        except UnboundLocalError:
            pass

        print("Waiting for sim to reply")
        print(env.last_action)


def show_video(frames: Sequence[np.ndarray], fps: int = 10, depth: bool = False):
    """Show a video composed of a sequence of frames.

    Example:
    frames = [
        controller.step("RotateRight", degrees=5).frame
        for _ in range(72)
    ]
    show_video(frames, fps=5)
    """
    rgb_frames = deepcopy(frames)
    if depth:
        rgb_frames = []
        for frame in frames:
            frame = frame.astype(np.float32)
            frame = (frame / (frame.max() + 1e-8)) * 255.0
            frame = frame.astype(np.uint8)
            frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)
            rgb_frames.append(frame)

    from moviepy import ImageSequenceClip

    frames = ImageSequenceClip(rgb_frames, fps=fps)
    return frames


def export_video(path, frames, depth: bool = False):
    """Merges all the saved frames into a .mp4 video and saves it to `path`"""

    video = cv2.VideoWriter(
        path,
        cv2.VideoWriter_fourcc(*"mp4v"),
        5,
        (frames[0].shape[1], frames[0].shape[0]),
    )
    for frame in frames:
        if depth:
            frame = frame.astype(np.float32)
            frame = (frame / (frame.max() + 1e-8)) * 255.0
            frame = frame.astype(np.uint8)
            frame_color = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)
            video.write(frame_color)
        else:
            video.write(cv2.cvtColor(frame, cv2.COLOR_RGB2BGR))
    cv2.destroyAllWindows()
    video.release()


def compute_rotate_camera_pose(center, pose, degree_per_frame=6):
    """degree_per_frame: set the degree of rotation for each frame"""

    def rotate_pos(x1, y1, x2, y2, degree):
        angle = math.radians(degree)
        n_x1 = (x1 - x2) * math.cos(angle) - (y1 - y2) * math.sin(angle) + x2
        n_y1 = (x1 - x2) * math.sin(angle) + (y1 - y2) * math.cos(angle) + y2

        return n_x1, n_y1

    x, z = rotate_pos(pose["position"]["x"], pose["position"]["z"], center["x"], center["z"], degree_per_frame)
    pose["position"]["x"], pose["position"]["z"] = x, z

    direction_x = center["x"] - x
    direction_z = center["z"] - z
    pose["rotation"]["y"] = math.degrees(math.atan2(direction_x, direction_z))

    return pose


def initialize_side_camera_pose(scene_bound, pose, third_fov=60, slope_degree=45, down_angle=70, scale_factor=8):
    """
    down_angle: the x-axis rotation angle of the camera, represents the top view of the front view from top to bottom, which needs to be less than 90 degrees
    ensure the line vector between scene's center & camera 's angel equal down_angle
    scale_factor scale the camera's view, make it larger ensure camera can see the whole scene
    """
    fov_rad = np.radians(third_fov)
    pitch_rad = np.radians(down_angle)
    distance = (scene_bound["center"]["y"] / 2) / np.tan(fov_rad / 2)
    pose["position"]["y"] = scene_bound["center"]["y"] + distance * scale_factor * np.sin(pitch_rad)
    pose["position"]["z"] = scene_bound["center"]["z"] - distance * scale_factor * np.cos(pitch_rad)

    pose["rotation"]["x"] = down_angle

    pose["orthographic"] = False
    del pose["orthographicSize"]

    pose = compute_rotate_camera_pose(scene_bound["center"], pose, slope_degree)

    return pose


def compute_timespan(timespan_str: str, dt: float) -> int:
    num, unit = timespan_str.split()
    num = int(num)
    unit = TIME_SCALES_MAPPING[unit]
    timespan = num * unit * (1 / dt)
    return int(timespan)


def run_keyboard_simulator(
    scene="FloorPlan205_physics",
    im_width: int = 640,
    im_height: int = 480,
    fov: int = 90,
    mindist_to_walls: float = 0.3,
    mindist_to_obj: float = 0.01,
    cleanup_previous_data: bool = False,
    generate_video: bool = False,
    generate_gif: bool = False,
    save_topdown: bool = True,
    batch_size: int = 128,
):


    sim = get_sim(
        scene,
        renderInstanceSegmentation=True,
        renderDepthImage=True,
        im_width=im_width,
        im_height=im_height,
        fov=fov,
        save_topdown=save_topdown,
        mindist_to_walls=mindist_to_walls,
        mindist_to_obj=mindist_to_obj,
    )
    controller = sim.controller

    sim.render()

    ## use keyboard control agent
    keyboard_play(
        controller,
        sim,
    )

    res_path = os.path.join(os.path.dirname(scene), f"privileged")
    save_sim_results(sim, res_path, batch_size, cleanup_previous_data, generate_video, generate_gif)


def run_simulator(
    scene="FloorPlan205_physics",
    im_width: int = 640,
    im_height: int = 480,
    fov: int = 90,
    mindist_to_walls: float = 0.3,
    mindist_to_obj: float = 0.01,
    timespan: str = "1 week",
    start_at: str = "0 hour",
    cleanup_previous_data: bool = False,
    generate_video: bool = False,
    generate_gif: bool = False,
    save_topdown: bool = True,
    batch_size: int = 128,
):

    # Load groundtruth config
    with open(os.path.join(scene, "config.yaml"), "r") as f:
        cfg = yaml.safe_load(f)

    gt_timespan = compute_timespan(cfg["timespan_to_collect"], cfg["dt"])
    timespan = compute_timespan(timespan, cfg["dt"])
    start_at = compute_timespan(start_at, cfg["dt"])
    assert timespan <= (
        gt_timespan - start_at
    ), f"Requested timespan {timespan} exceeds collected timespan {gt_timespan} when starting at {start_at}"

    sim = get_sim(
        scene,
        renderInstanceSegmentation=True,
        renderDepthImage=True,
        im_width=im_width,
        im_height=im_height,
        fov=fov,
        save_topdown=save_topdown,
        mindist_to_walls=mindist_to_walls,
        mindist_to_obj=mindist_to_obj,
    )
    # Automatically navigate to all receptacles
    sim.kill_sim_on_condition_failure = False
    sim.raise_exception_on_condition_failure = False

    # Tick the clock to the start time
    for n_ticks in range(start_at):
        # At the last iteration step objects without rendering
        if n_ticks == (start_at - 1):
            sim.StepSemiStaticObjects() # render=False)
        else:
            sim.sss_data = sim.sss_data.step()
            sim.privileged_apn = None  # resets the accumulator

    sim.render()
    print("**************************************")
    print(f"Starting collecting data at major tick {start_at}")
    print("**************************************")

    try:
        for i in range(timespan):
            print(f"Navigating sequence at time step {i + 1} / {timespan}")
            sim.GoToAllInstances(ROBOTS[0])
            sim.privileged_apn = sim.privileged_apn.backprop_majors()
            sim.privileged_apn = sim.privileged_apn.dump_to_parquet(
                sim.tmpdir, dump_leftover=False, verbose=True, batch_size=batch_size
            )
            # Once done navigating to all receptacles, step semi-static objects and teleport
            random_teleport(sim.controller, sim.full_reachability_graph)
            sim.StepSemiStaticObjects()
            sim.render()
            sim.privileged_apn = sim.privileged_apn.remove_nan_sentinel()

        print("**************************************")
        print("Finished navigating to all receptacles")
        print("**************************************")
    except KeyboardInterrupt:
        print("\n!!! User interrupted simulation (Ctrl+C). Saving partial data... !!!")
    except Exception as e:
        raise e
        print(f"\n!!! Simulation encountered an error: {e}. Saving partial data... !!!")
    finally:
        print("Saving simulation results...")
        res_path = os.path.join(os.path.dirname(scene), f"privileged")
        # poke
        # sim.privileged_apn.sum_up_minors_into_major()
        save_sim_results(sim, res_path, batch_size, cleanup_previous_data, generate_video, generate_gif)


@hydra.main(version_base=None, config_path="config", config_name="config")
def _main(cfg: DictConfig):
    import os

    os.environ["JAX_PLATFORM_NAME"] = "cpu"
    import jax

    jax.config.update("jax_platform_name", "cpu")

    hydra.utils.call(cfg.mode)


def main(version_base=None, config_path="config", config_name="config"):
    try:
        _main(version_base=None, config_path="config", config_name="config")
    except Exception as e:
        print(f"An error occurred: {e}")
        print("Attempting to fall back to default config")
        return _main(
            version_base=None,
            config_path=os.path.join(os.path.dirname(__file__), "config"),
            config_name="config",
        )

if __name__ == "__main__":
    main()



import networkx as nx

from semistaticsim.rendering.floor import ReachabilityGraph
from semistaticsim.rendering.utils.dict2xyztup import dict2xyztuple


def cleanup_request(requested_node):
    if isinstance(requested_node, dict) and all([x in requested_node for x in ['x', 'y', 'z']]):
        requested_node = dict2xyztuple(requested_node)
    if isinstance(requested_node, tuple):
        requested_node = list(requested_node)
    return requested_node

def astar(requested_start_node, requested_end_node, filtered_rg: ReachabilityGraph, controller):
    requested_start_node = cleanup_request(requested_start_node)
    requested_end_node = cleanup_request(requested_end_node)

    # filtered_rg = full_reachability_graph.filter_(controller.last_event.metadata["objects"], mindist_to_obj=0.5)

    # filtered_rg.parent_floor.plot_self(reachability_graph=filtered_rg)
    # filtered_rg.parent_floor.plot_self(reachability_graph=filtered_rg)
    # filtered_rg.parent_floor.plot_self(samples=jnp.array(path))

    start_node = filtered_rg.shunt(requested_start_node)
    end_node = filtered_rg.shunt(requested_end_node)

    assert start_node in filtered_rg.full_graph
    assert end_node in filtered_rg.full_graph

    path = nx.astar_path(filtered_rg.full_graph, start_node, end_node)[1:]


    for node in path:
        yield node



import shutil
import numpy as np
import os

from pathlib import Path

from tqdm import tqdm

from semistaticsim.datawrangling.sssd import load_sssd, GeneratedSemiStaticData


def purge_dir(res_path, prefix):
    for item in os.listdir(res_path):
        item_path = os.path.join(res_path, item)
        if not os.path.isdir(item_path) or not item.startswith(prefix):
            if os.path.isdir(item_path):
                shutil.rmtree(item_path)
            else:
                os.remove(item_path)


def move_indexed_files(src_dir: Path, dst_dir: Path, indices, ext):
    dst_dir.mkdir(parents=True, exist_ok=True)
    for trg_idx, src_idx in enumerate(indices):
        shutil.move(src_dir / f"{src_idx:05d}.{ext}", dst_dir / f"{trg_idx:05d}.{ext}")


def save_run(mask: np.ndarray, sssd_data: GeneratedSemiStaticData, res_path: str, tmpdir: str, batch_size: int):
    # Delete existing results
    if os.path.exists(res_path):
        shutil.rmtree(res_path)
    os.makedirs(res_path, exist_ok=True)
    # Index df samples
    indices = np.nonzero(mask)[0]
    subset = sssd_data.take_by_global_timestep(stop_or_start=None, indices=indices)
    # --- Move Images ---
    images_src = tmpdir / "images"
    images_dst = res_path / "images"
    for subdir in images_src.iterdir():
        if subdir.is_dir():
            # detect extension from first file
            sample = next(subdir.iterdir())
            ext = sample.suffix.lstrip(".")

            move_indexed_files(src_dir=subdir, dst_dir=images_dst / subdir.name, indices=indices, ext=ext)

    # --- Move poses ---
    move_indexed_files(src_dir=tmpdir / "poses", dst_dir=res_path / "poses", indices=indices, ext="json")

    # Save parquet
    subset.dump_to_parquet(res_path, dump_leftover=True, verbose=False, batch_size=batch_size)

    # Copy id_to_color mapping if exists
    id_to_color_src = tmpdir / "id_to_color.json"
    if id_to_color_src.exists():
        shutil.copy(id_to_color_src, res_path / "id_to_color.json")


def format_results(res_path: str, tmpdir: str, batch_size: int) -> dict:
    sssd_data = load_sssd(tmpdir)
    # Get timestamps to split data folder
    timestamps = np.concatenate([data._timestamp for data in sssd_data.get_generator_of_selves()])
    floored_timestamps = np.floor(timestamps)
    unique_timestamps = np.unique(floored_timestamps)
    # Split data into multiple folders based on timestamps
    for t in tqdm(unique_timestamps, desc="Formatting results"):
        mask = floored_timestamps == t
        if mask.sum() > 1:
            print(f"Processing run_id: {int(t.item())} with {mask.sum()} samples")
            save_run(mask, sssd_data, res_path=Path(res_path) / f"run_{int(t.item())}", tmpdir=Path(tmpdir), batch_size=batch_size)


import dataclasses
import functools
import json
import os
import re
from typing import List, Optional, Tuple, Dict

import yaml
from typing_extensions import Self

import jax
import jax.random
import numpy as np
import polars as pl
from flax import struct
from jax import numpy as jnp
from pathlib import Path


def thin2wide(assignment, receptacles_in_scene):
    def do_for_each_p(r_or_n_value):
        mat = jnp.zeros(
            (
                len(
                    receptacles_in_scene,
                )
            )
        )
        mat = mat.at[r_or_n_value].set(1)
        return mat

    def for_each_timestep(p_r_apn):
        return jax.vmap(do_for_each_p)(p_r_apn)

    assignment = assignment.astype(int)
    return jax.vmap(for_each_timestep)(assignment)


def path_2_parts(path: str | Path) -> Tuple[str, int, jax.random.PRNGKey]:
    p = Path(path).resolve()
    parts = list(p.parts)
    for i in range(len(parts) - 2):
        if parts[i].isalpha() and parts[i + 1].isdigit() and parts[i + 2].isdigit():
            procthor_split = parts[i]
            procthor_index = int(parts[i + 1])
            jax_seed = int(parts[i + 2])
            return procthor_split, procthor_index, jax_seed 


@struct.dataclass
class GeneratedSemiStaticData:
    split: str
    house_id: str
    jax_key: jax.random.PRNGKey  # seed
    dt: float

    _assignment: jnp.ndarray
    _timestamp: jnp.ndarray
    _aabb_center: jnp.ndarray
    _aabb_cornerPoints: jnp.ndarray
    _aabb_size: jnp.ndarray
    _oobb_cornerPoints: jnp.ndarray
    _position: jnp.ndarray
    _rotation: jnp.ndarray

    pickupable_names: List[str]
    receptacle_names: List[str]
    pickupable_to_receptacle: Dict[str, str]
    receptacles_oobb: Dict[str, Dict]

    is_single_timestamp: bool  # Is self sliced at a timestamp (single timestamp strict)?
    is_single_pickupable: (
        bool  # Is self sliced at a pickupable (single pickupable strict and also single timestamp strict)?
    )

    parquet_paths: List[str]

    intrinsics: Optional[Dict[str, Dict]] = None
    parquet_id: int = 0
    timestep: int = 0  # Which timestep index are we at
    parent: Self = None  # Self but with all info if sliced, otherwise none

    def __getitem__(self, item):
        """Index class attributes by string key"""
        if isinstance(item, str):
            return dataclasses.asdict(self)[item]
        raise NotImplementedError()

    def get_receptacle_oobb(self, receptacle_name) -> Dict[str, Dict]:
        """Get the receptacle OOBB information"""
        return self.receptacles_oobb[receptacle_name]

    @property
    def original_batch_size(self) -> int:
        """What is the parquets' batch size"""
        return self._resolve_original_parent()._timestamp.shape[0]

    @property
    def global_timestep(self) -> int:
        """What is the global timestep index across all parquets"""
        return self.original_batch_size * self.parquet_id + self.timestep

    @property
    def global_length(self):
        """Total number of timesteps across all parquets"""

        if len(self.parquet_paths) == 0:
            return self._timestamp.shape[0]

        temp_self = self.replace(parquet_id=len(self.parquet_paths) - 1).resolve_parquet()
        return (len(self.parquet_paths) - 1) * self.original_batch_size + temp_self._timestamp.shape[0]

    def _resolve_original_parent(self) -> Self:
        """Recursively resolve to the original parent if sliced, otherwise return self"""
        if self.parent is None:
            return self
        return self.parent._resolve_original_parent()

    def take_by_global_timestep(self, stop_or_start, stop=None, step=None, indices=None) -> Self:
        """Index a slice of data across all parquets
        self.take_by_global_timestep(20, 50)  # Take data from index 20 (inclusive) to 50 (non inclusive) across all parquets
        self.take_by_global_timestep(None)    # Takes all data from all parquets
        self.take_by_global_timestep(None, indices=[0, 5, 10, 15])  # Take specific indices
        """
        if indices is not None:
            # Use the provided indices directly
            indices_to_take = jnp.asarray(indices).squeeze()
        else:
            # Use slice notation
            data_slice = slice(stop_or_start, stop, step)
            indices_to_take = jnp.arange(self.global_length)[data_slice].squeeze()

        parquet_id_to_timestep_id = {}
        og_batch_size = self.original_batch_size
        for i in indices_to_take:
            parquet_id = int(i / og_batch_size)
            timestep_id = int(i - parquet_id * og_batch_size)

            if parquet_id not in parquet_id_to_timestep_id:
                parquet_id_to_timestep_id[parquet_id] = []
            parquet_id_to_timestep_id[parquet_id].append(timestep_id)
        parquet_id_to_timestep_id = {k: jnp.array(v) for k, v in parquet_id_to_timestep_id.items()}

        ret = []
        for parquet_id, timestep_ids in parquet_id_to_timestep_id.items():
            self = self.replace(parquet_id=parquet_id, timestep=0).resolve_parquet()

            ret.append(self.replace(**{k: v[timestep_ids] for k, v in self.TimeVaryingItems()}))

        ret_self = ret[0]
        for self in ret[1:]:
            ret_self = ret_self.concat(self)
        return ret_self

    @property
    def pickupables_in_scene(self):
        """Get all pickupable names in the scene - same as pickupable_names.json"""
        return self._resolve_original_parent().pickupable_names

    @property
    def receptacles_in_scene(self):
        """Get all receptacle names in the scene - same as receptacle_names.json"""
        return self._resolve_original_parent().receptacle_names

    @classmethod
    def TimeVaryingKeys(cls) -> List[str]:
        """Returns the name of the keys that are in the parquet files (time varying)"""
        return [f.name for f in dataclasses.fields(cls) if f.name.startswith("_")]

    def TimeVaryingItems(self) -> dict.items:
        """Returns the items (key, value) of the time varying keys (parquets)"""
        asdict = dataclasses.asdict(self)
        return {k: v for k, v in asdict.items() if k.startswith("_")}.items()

    @property
    def is_full_timestamps(self) -> bool:
        """Check if current self is either a single timestamp or full data"""
        return not self.is_single_timestamp

    @property
    def is_full_pickupables(self) -> bool:
        """Check if current self is either a single pickupable or full data"""
        return not self.is_single_pickupable

    def unstep(self) -> Self:
        """
        Step to the next timestep, or next parquet if at the end of current parquet

        NOTE: Used mainly during data generation
        """
        self = self.replace(timestep=self.timestep - 1)
        if self.timestep < 0:
            if self.parquet_id > 0:
                self = self.replace(parquet_id=self.parquet_id - 1, timestep=self.original_batch_size-1)
                self = self.resolve_parquet()
            else:
                print("WARNING: all generated semi static object steps are DONE!")
                return None
        return self

    def step(self) -> Self:
        """
        Step to the next timestep, or next parquet if at the end of current parquet

        NOTE: Used mainly during data generation
        """
        self = self.replace(timestep=self.timestep + 1)
        if self.timestep >= self._assignment.shape[0]:
            if self.parquet_id < len(self.parquet_paths):
                self = self.replace(parquet_id=self.parquet_id + 1, timestep=0)
                self = self.resolve_parquet()
            else:
                print("WARNING: all generated semi static object steps are DONE!")
                return None
        return self

    def get_generator_of_selves(self):
        """Iterator that returns a (full) self for each parquet file"""
        yield self
        while self.parquet_id < len(self.parquet_paths) - 1:
            self = self.replace(parquet_id=self.parquet_id + 1, timestep=0)
            self = self.resolve_parquet()
            yield self

    def resolve_parquet(self) -> Self:
        """Update self by loading data from the parquet file at self.parquet_id - same as read_parquet"""
        ITEMS = {key: None for key in self.TimeVaryingKeys()}

        path = self.parquet_paths[self.parquet_id]
        df = pl.read_parquet(path)

        for key in self.TimeVaryingKeys():
            ITEMS[key] = jnp.array(df[key[1:]]).squeeze()

        if len(ITEMS["_assignment"].shape) == 2:
            # edge case: the parquet contains a single datapoint
            for key in self.TimeVaryingKeys():
                ITEMS[key] = ITEMS[key][None]

        #if len(ITEMS["_assignment"].shape) == 2:
        #    ITEMS["_assignment"] = thin2wide(ITEMS["_assignment"], self.receptacles_in_scene)

        self = self.replace(**ITEMS)
        return self

    @property
    def self_at_current_time(self) -> Self:
        """Slice self at current timestep index __get_item__(timestamp)
        where it retusn self as opposed to an element in the container"""
        return self.replace(
            **{k: v[self.timestep] for k, v in self.TimeVaryingItems()}, is_single_timestamp=True, parent=self
        )

    @property
    def pickupable_selves_at_current_time(self) -> List[Self]:
        """Returns a dictionary pickupable_name -> self sliced at current timestep and pickupable"""
        if not self.is_single_timestamp:
            current_time_self = self.self_at_current_time

        ret = {}
        for i, p in enumerate(self.pickupables_in_scene):
            new_self = current_time_self.replace(
                pickupable_names=[p],
                receptacle_names=self.pickupable_to_receptacle[p],
                **({k: (v[i] if "timestamp" not in k else v) for k, v in current_time_self.TimeVaryingItems()}),
                is_single_pickupable=True,
                parent=current_time_self,
            )
            ret[p] = new_self
        return ret

    def current_receptacle_for_this_pickupable(self, pickupable_name: str) -> str:
        """Which receptacle is the given pickupable currently in? If not returns unobserved message"""
        if self.is_single_pickupable:
            assert self.is_single_timestamp
            assignment = jnp.atleast_1d(self._assignment).argmax()
            return (
                self.receptacles_in_scene[assignment]
                if self._assignment[assignment] == 1
                else f"Pickupable {pickupable_name} was unobserved"
            )
        return self.pickupable_selves_at_current_time[pickupable_name].current_receptacle_for_this_pickupable(
            pickupable_name
        )

    def is_this_pickupable_in_the_OOB_FAKE_RECEPTACLE(self, pickupable_name: str) -> bool:
        """Check if a given pickupable is currently in the OOB_FAKE_RECEPTACLE"""
        return self.current_receptacle_for_this_pickupable(pickupable_name) == "OOB_FAKE_RECEPTACLE"

    def get_singletimestamp_prototype(self) -> Self:
        """
        Get a prototype self that is single timestamp with all data as NaN
        """
        kwargs = {}
        for k, v in self.self_at_current_time.TimeVaryingItems():
            if "assignment" in k:
                kwargs[k] = jnp.ones((len(self.pickupables_in_scene), len(self.receptacles_in_scene))) * jnp.nan
            else:
                kwargs[k] = jnp.ones_like(v) * jnp.nan
        return self.replace(
            **kwargs,
            parent=None,
            is_single_timestamp=True,
            is_single_pickupable=False,
            pickupable_names=self.pickupables_in_scene,
            receptacle_names=self.receptacles_in_scene,
            parquet_paths=[],
            parquet_id=0
        )

    def concat(self, other_self: Self):
        """Concatenate two selves along the timestamp dimension"""
        # assumes self to be the accumulator
        assert not self.is_single_pickupable

        if self.is_single_timestamp:
            # the current self is a point
            self = self.replace(**{k: v[None] for k, v in self.TimeVaryingItems()}, is_single_timestamp=False)

        kwargs = {}
        for k, v in self.TimeVaryingItems():
            if other_self.is_single_timestamp:
                kwargs[k] = jnp.concatenate([v, other_self[k][None]])
            else:
                kwargs[k] = jnp.concatenate([v, other_self[k]])

        concatenated_self = self.replace(**kwargs)
        return concatenated_self

    def remove_nan_sentinel(self):
        if jnp.isnan(self._timestamp[0]).any():
            self = self.replace(**{k: v[1:] for k, v in self.TimeVaryingItems()})
            assert not jnp.isnan(self._timestamp[0]).any()
        return self

    def backprop_majors(self):
        UNIQUE_TIMESTAMPS, UNIQUE_TIMESTAMP_INDEXES, UNIQUE_TIMESTAMP_COUNTS = jnp.unique(
            self._timestamp, return_counts=True, return_index=True
        )
        UNIQUE_TIMESTAMPS = list(map(float, UNIQUE_TIMESTAMPS))

        self_summed_majors = self.sum_up_minors_into_major(verbose=True)
        for i, (unique_timestamp, unique_timestamp_index, unique_timestamp_count) in enumerate(zip(UNIQUE_TIMESTAMPS, UNIQUE_TIMESTAMP_INDEXES, UNIQUE_TIMESTAMP_COUNTS)):

            major_index = jnp.argmax(self_summed_majors._timestamp == int(unique_timestamp))
            major_assignment = self_summed_majors._assignment[major_index]

            if i + 1 < len(UNIQUE_TIMESTAMP_INDEXES):
                slice = self._assignment.at[UNIQUE_TIMESTAMP_INDEXES[i] : UNIQUE_TIMESTAMP_INDEXES[i + 1]]
            else:
                slice = self._assignment.at[UNIQUE_TIMESTAMP_INDEXES[i] :]

            sliced = slice.get()
            masked = jnp.astype(sliced >= 0, int) # indices all timestamps where we have non-missing data for the receptacles

            new_sliced = jax.vmap(lambda s, m: major_assignment * m + s * (1-m))(sliced, masked)
            #new_sliced = major_assignment * masked + sliced * (1-masked)

            new_self = self.replace(_assignment=slice.set(new_sliced))
            assert new_self._assignment.shape == self._assignment.shape
            self = new_self
            del new_self

        return self

    def regress_minors_back_into_raw_minors(self):
        return self.replace(_timestamp=jnp.astype(jnp.astype(self._timestamp, int), float))

    def sum_up_minors_into_major(self, verbose=True):
        if verbose:
            print("Summing up minors into major. This only works if you call it right after a major loop has finished! Otherwise you might introduce bugs in the saved data.")

        assert not self.is_single_pickupable
        assert not self.is_single_timestamp

        self = self.regress_minors_back_into_raw_minors()

        UNIQUE_TIMESTAMPS, UNIQUE_TIMESTAMP_INDEXES, UNIQUE_TIMESTAMP_COUNTS = jnp.unique(
            self._timestamp, return_counts=True, return_index=True
        )
        UNIQUE_TIMESTAMPS = list(map(float, UNIQUE_TIMESTAMPS))

        PROTO = self.get_singletimestamp_prototype()
        MAJOR_TO_PROTO = {}
        for i, (value, index, count) in enumerate(zip(UNIQUE_TIMESTAMPS, UNIQUE_TIMESTAMP_INDEXES, UNIQUE_TIMESTAMP_COUNTS)):
            if i + 1 < len(UNIQUE_TIMESTAMP_INDEXES):
                slice = self._assignment.at[UNIQUE_TIMESTAMP_INDEXES[i] : UNIQUE_TIMESTAMP_INDEXES[i + 1]]
            else:
                slice = self._assignment.at[UNIQUE_TIMESTAMP_INDEXES[i] :]
            slice = slice.get()

            MAJOR_FOR_Ps = []
            for p_id, p_name in enumerate(self.pickupables_in_scene):
                minor_assignments_for_p = slice[:, p_id, :]
                major_assignments_for_p = jnp.max(minor_assignments_for_p, axis=0)

                for r_id, r_name in enumerate(self.receptacles_in_scene):
                    if r_name not in self.pickupable_to_receptacle[p_name]:
                        major_assignments_for_p = major_assignments_for_p.at[r_id].set(-2)

                major_assignments_for_p = major_assignments_for_p[None,None,:]
                MAJOR_FOR_Ps.append(major_assignments_for_p)

            major = jnp.concatenate(MAJOR_FOR_Ps, axis=1)
            self_at_this_time = self.replace(timestep=UNIQUE_TIMESTAMP_INDEXES[i]).self_at_current_time
            MAJOR_TO_PROTO[value] = PROTO.replace(
                _assignment=major.astype(int),
                _timestamp=jnp.array([value]),
                **{k: v[None] for k,v in self_at_this_time.TimeVaryingItems() if (k != "_assignment" and k != "_timestamp")},
                is_single_timestamp=False
            )

        result = MAJOR_TO_PROTO[UNIQUE_TIMESTAMPS[0]]
        # fixme is this correct behaviour: throws away the last one because the last one should have no minors (?)
        for other_self in [MAJOR_TO_PROTO[val] for val in UNIQUE_TIMESTAMPS[1:]]:
            result = result.concat(other_self)
        return result

    def process_raw_minors(self):
        UNIQUE_TIMESTAMPS, UNIQUE_TIMESTAMP_INDEXES, UNIQUE_TIMESTAMP_COUNTS = jnp.unique(
            self._timestamp, return_counts=True, return_index=True
        )
        if jnp.all(UNIQUE_TIMESTAMP_COUNTS == 1):
            print("Not doing the sub_timestamp computation because there is no need: all timestamps are unique!")
        else:
            for i, (index, count) in enumerate(zip(UNIQUE_TIMESTAMP_INDEXES, UNIQUE_TIMESTAMP_COUNTS)):
                if count == 1:
                    continue

                if i + 1 < len(UNIQUE_TIMESTAMP_INDEXES):
                    slice = self._timestamp.at[UNIQUE_TIMESTAMP_INDEXES[i] : UNIQUE_TIMESTAMP_INDEXES[i + 1]]
                else:
                    slice = self._timestamp.at[UNIQUE_TIMESTAMP_INDEXES[i] :]

                sliced = slice.get()

                if sliced.shape[0] == 1:
                    continue

                SHAPE_PAD = sliced.shape[0] + 1
                SUB_TIMESTAMP = self.dt / SHAPE_PAD
                incremental = SUB_TIMESTAMP * jnp.arange(SHAPE_PAD)[1:]
                self = self.replace(_timestamp=slice.set(sliced + incremental))
        return self

    def dump_to_parquet(self, target_dir, dump_leftover=False, batch_size: int = 100, verbose=True) -> Self:
        """
        Dumps current self to parquet
        Args:
            target_dir:
            batch_size:
            dump_leftover: whether to dump the final leftover data that doesn't fit in a batch_size


        Returns:
            leftover data that didn't fit in a batch_size
        """
        if verbose:
            print("DUMPING TO PARQUET. Note: this function only works ONCE! Call it at the end of data generation.")
            print(
                "To make it work for multiple dumping episodes, you need to implement a different sub_timestamp logic:"
            )
            print("right now, it only computes the sub_timestamps by counting all consecutive identical timestamps and")
            print(
                "cumulatively adding 1/NUM_IDENTICAL_TIMESTAMPS to simulate the time taken to navigate the room by the agent"
            )

        self = self.remove_nan_sentinel()
        self = self.process_raw_minors()

        num_full_batches = self._timestamp.shape[0] // batch_size
        leftover_data = self._timestamp.shape[0] - num_full_batches * batch_size

        # todo write some code to find the next id to write in os.path.join(target_dir, f"scan_{i}.parquet"))
        existing = [f for f in os.listdir(target_dir) if f.startswith("scan_") and f.endswith(".parquet")]
        existing_ids = [
            int(re.search(r"scan_(\d+)\.parquet", f).group(1)) for f in existing if re.search(r"scan_(\d+)\.parquet", f)
        ]
        CUR_SCAN_ID = max(existing_ids, default=-1) + 1  # start after the last one

        CUR_IDX = 0
        WROTE_DATA = False
        for batch_id in range(num_full_batches):
            kwargs = {k.lstrip("_"): np.array(v)[CUR_IDX : CUR_IDX + batch_size] for k, v in self.TimeVaryingItems()}
            CUR_IDX += batch_size
            df = pl.DataFrame(kwargs)
            df.write_parquet(os.path.join(target_dir, f"scan_{CUR_SCAN_ID}.parquet"))
            CUR_SCAN_ID += 1
            WROTE_DATA = True

        if leftover_data >= 0 and dump_leftover:
            leftover_data_kwargs = {k.lstrip("_"): np.array(v)[CUR_IDX:] for k, v in self.TimeVaryingItems()}
            df = pl.DataFrame(leftover_data_kwargs)
            df.write_parquet(os.path.join(target_dir, f"scan_{CUR_SCAN_ID}.parquet"))
            WROTE_DATA = True

        if WROTE_DATA:
            with open(os.path.join(target_dir, "pickupable_names.json"), "w") as f:
                json.dump(self.pickupable_names, f, indent=4)
            with open(os.path.join(target_dir, "receptacle_names.json"), "w") as f:
                json.dump(self.receptacle_names, f, indent=4)
            with open(os.path.join(target_dir, "pickupable_to_receptacle.json"), "w") as f:
                json.dump(self.pickupable_to_receptacle, f, indent=4)
            with open(os.path.join(target_dir, "receptacles_oobb.json"), "w") as f:
                json.dump(self.receptacles_oobb, f, indent=4)
            with open(os.path.join(target_dir, "config.yaml"), "w") as f:
                yaml.dump({"dt": float(self.dt)}, f)
            if self.intrinsics is not None:
                with open(os.path.join(target_dir, "camera_intrinsics.json"), "w") as f:
                    json.dump(self.intrinsics, f, indent=4)

        if not dump_leftover:
            if leftover_data > 0:
                return self.replace(
                    **{k: v[CUR_IDX:] for k, v in self.TimeVaryingItems()}
                )
            else:
                return self.replace(
                    **{k: jnp.zeros_like(v[CUR_IDX-1:][0][None]) * jnp.nan for k, v in self.TimeVaryingItems()}
                )
        return None

def load_sssd(path):
    # Load a config file
    split, house_id, jax_key = path_2_parts(path)

    ITEMS = {key: None for key in GeneratedSemiStaticData.TimeVaryingKeys()}

    def extract_index(name: str) -> int:
        base = name.replace(".parquet", "")
        parts = base.split("_")
        if len(parts) == 1:
            return 0
        return int(parts[-1])

    file_names = [f for f in os.listdir(path) if f.endswith(".parquet")]
    file_names = sorted(file_names, key=extract_index)
    file_names = [os.path.join(path, file) for file in file_names]

    with open(os.path.join(path, "pickupable_names.json"), "r") as f:
        pickupables = json.load(f)
    with open(os.path.join(path, "receptacle_names.json"), "r") as f:
        receptacles = json.load(f)
    with open(os.path.join(path, "pickupable_to_receptacle.json"), "r") as f:
        pickupable_to_receptacle = json.load(f)
    with open(os.path.join(path, "receptacles_oobb.json"), "r") as f:
        receptacles_oobb = json.load(f)
    intrinsics = None
    intrinsics_path = os.path.join(path, "camera_intrinsics.json")
    if os.path.exists(intrinsics_path):
        with open(intrinsics_path, "r") as f:
            intrinsics = json.load(f)

    try:
        with open(os.path.join(path, "config.yaml"), "r") as f:
            config = yaml.safe_load(f)
            dt = config["dt"]
    except:
        raise Exception("No config.yaml found in the SSSD path, this should ALWAYS exist!")

    return GeneratedSemiStaticData(
        split=split,
        house_id=house_id,
        jax_key=jax.random.PRNGKey(jax_key),
        dt=dt,
        parquet_paths=file_names,
        **ITEMS,
        pickupable_names=pickupables,
        receptacle_names=receptacles,
        pickupable_to_receptacle=pickupable_to_receptacle,
        receptacles_oobb=receptacles_oobb,
        intrinsics=intrinsics,
        is_single_timestamp=False,
        is_single_pickupable=False,
        # batch_size=cfg['jax_scan_size']
    ).resolve_parquet()


if __name__ == "__main__":
    import os

    os.environ["JAX_PLATFORM_NAME"] = "cpu"
    import jax

    jax.config.update("jax_platform_name", "cpu")

    x = load_sssd("../generated_data/semistaticsim/test/0/0/privileged/run_2")
    print({k:v.shape for k,v in x.take_by_global_timestep(None).TimeVaryingItems()})
    #print(x.get_current_pickupable_posrot(name=x.pickupable_names[0]))

"""

    ### ASSIGNMENT CONVERSION ###
    def convert_to_wide_assignment_tensor(self):
        if self.is_wide:
            return self

        assert not jnp.any(self._assignment == -1), "Thin to wide requires no missing data"

        def do_for_each_p(r_or_n_value):
            mat = jnp.zeros((self.num_receptacles_in_scene,))
            mat = mat.at[r_or_n_value].set(1)
            return mat

        if self.is_single_pickupable:
            assert self.is_single_timestamp
            return self.replace(_assignment=do_for_each_p(self._assignment))

        if self.is_single_timestamp:
            return self.replace(_assignment=jax.vmap(do_for_each_p)(self._assignment))

        def for_each_timestep(p_r_apn):
            return jax.vmap(do_for_each_p)(p_r_apn)
        return self.replace(_assignment=jax.vmap(for_each_timestep)(self._assignment))
"""



def xyztuple2dict(tup):
    return {'x': tup[0], 'y': tup[1], 'z': tup[2]}

def dict2xyztuple(dic):
    return (dic['x'], dic['y'], dic['z'])

def xyztuple_precision(tup):
    return (round(tup[0], 3), round(tup[1],3), round(tup[2],3))


import os

import jax.numpy as jnp

os.environ["JAX_PLATFORM_NAME"] = "cpu"
import jax
jax.config.update('jax_platform_name', "cpu")


def spoof_vmap(list_of_dataclasses):
    leaves_proto, tree_def = jax.tree.flatten(list_of_dataclasses[0])
    leaves_acc = []
    for leaf in leaves_proto:
        if isinstance(leaf, jnp.ndarray):
            pass
        else:
            leaf = jnp.array([leaf])
        leaves_acc.append(leaf[None])

    for i, o in enumerate(list_of_dataclasses[1:]):
        leaves, tree_def = jax.tree.flatten(o)
        if leaves_acc is None:
            leaves_acc = leaves
        else:
            tmp = []
            for leaf_acc, leaf in zip(leaves_acc, leaves):
                if isinstance(leaf, jnp.ndarray):
                    pass
                else:
                    leaf = jnp.array([leaf])
                tmp.append(jnp.concatenate([leaf_acc, leaf[None]]))
            leaves_acc = tmp
    ret = jax.tree.unflatten(tree_def, leaves_acc)
    return ret